import unittest

import yara

from maco import extractor


class TestExtractor(unittest.TestCase):
    def test_bad(self):
        class Tmp(extractor.Extractor):
            family = "smell_ya_later"
            author = "me"
            last_modified = "yeah"

        Tmp()

        class Tmp1(Tmp):
            family = None

        self.assertRaises(extractor.InvalidExtractor, Tmp1)

        class Tmp1(extractor.Extractor):
            author = None

        self.assertRaises(extractor.InvalidExtractor, Tmp1)

        class Tmp1(extractor.Extractor):
            version = None

        self.assertRaises(extractor.InvalidExtractor, Tmp1)

        class Tmp1(Tmp):
            yara_rule: str = "t"

        self.assertRaises(extractor.InvalidExtractor, Tmp1)

        class Tmp1(Tmp):
            yara_rule = """
                rule DifferentName
                {
                    condition:
                        true
                }
            """

        Tmp1()

        class Tmp1(Tmp):
            yara_rule = """
                rule Tmp1
                {
                    condition:
                        true
                }
                rule OtherName
                {
                    condition:
                        true
                }
            """

        Tmp1()



"""CLI example of how extractors can be executed."""
import argparse
import base64
import binascii
import hashlib
import io
import json
import logging
import os
import sys
from typing import BinaryIO, List

import cart

from maco import collector

logger = logging.getLogger("maco.lib.cli")


def process_file(
    collected: collector.Collector,
    path_file: str,
    stream: BinaryIO,
    *,
    pretty: bool,
    force: bool,
    include_base64: bool,
):
    """Process a filestream with the extractors and rules."""
    unneutered = io.BytesIO()
    try:
        cart.unpack_stream(stream, unneutered)
    except Exception:
        # use original stream if anything goes wrong here
        # i.e. invalid/malformed cart
        pass
    else:
        # use unneutered stream
        stream = unneutered
    # unpack will read some bytes either way so reset position
    stream.seek(0)

    # find extractors that should run based on yara rules
    if not force:
        runs = collected.match(stream)
    else:
        # execute all extractors with no yara information
        # note - extractors may rely on a yara hit so this may cause errors
        runs = {x: [] for x in collected.extractors.keys()}
    if not runs:
        return

    # run extractor for the set of hits
    logger.info(f"path: {path_file}")
    ret = {}
    for extractor_name, hits in runs.items():
        # run and store results for extractor
        logger.info(f"run {extractor_name} extractor from rules {[x.rule for x in hits]}")
        try:
            resp = collected.extract(stream, hits, extractor_name)
        except Exception as e:
            logger.exception(f"extractor error with {path_file} ({e})")
            resp = None
        # encode binary data so we can print as json
        if resp:
            for row in resp.get("binaries", []):
                row["sha256"] = hashlib.sha256(row["data"]).hexdigest()
                # number of bytes in the binary
                row["size"] = len(row["data"])
                # small sample of first part of binary
                row["hex_sample"] = binascii.hexlify(row["data"][:32]).decode("utf8").upper()
                if include_base64:
                    # this can be large
                    row["base64"] = base64.b64encode(row["data"]).decode("utf8")
                # do not print raw bytes to console
                row.pop("data")
        ret[extractor_name] = resp
        logger.info(json.dumps(resp, indent=2 if pretty else None))
    logger.info("")

    return ret


def process_filesystem(
    path_extractors: str,
    path_samples: str,
    include: List[str],
    exclude: List[str],
    *,
    pretty: bool,
    force: bool,
    include_base64: bool,
    create_venv: bool = False,
):
    if force:
        logger.warning("force execute will cause errors if an extractor requires a yara rule hit during execution")
    collected = collector.Collector(path_extractors, include=include, exclude=exclude, create_venv=create_venv)

    logger.info(f"extractors loaded: {[x for x in collected.extractors.keys()]}\n")
    for _, extractor in collected.extractors.items():
        extractor = extractor["module"]
        logger.info(
            f"{extractor.family} by {extractor.author}"
            f" {extractor.last_modified} {extractor.sharing}"
            f"\n{extractor.__doc__}\n"
        )

    num_analysed = 0
    num_hits = 0
    num_extracted = 0
    if os.path.isfile(path_samples):
        # analyse a single file
        walker = [("", None, [path_samples])]
    elif os.path.isdir(path_samples):
        # load files from directory tree
        walker = os.walk(path_samples)
    else:
        logger.error(f"not file or folder: {path_samples}")
        exit(2)
    try:
        for path, dirs, files in walker:
            for file in files:
                num_analysed += 1
                path_file = os.path.join(path, file)
                try:
                    with open(path_file, "rb") as stream:
                        resp = process_file(
                            collected,
                            path_file,
                            stream,
                            pretty=pretty,
                            force=force,
                            include_base64=include_base64,
                        )
                        if resp:
                            num_hits += 1
                            if any(x for x in resp.values()):
                                num_extracted += 1
                except Exception as e:
                    logger.exception(f"file error with {path_file} ({e})")
                    continue
    except:
        raise
    finally:
        logger.info("")
        logger.info(f"{num_analysed} analysed, {num_hits} hits, {num_extracted} extracted")


def main():
    parser = argparse.ArgumentParser(description="Run extractors over samples.")
    parser.add_argument("extractors", type=str, help="path to extractors")
    parser.add_argument("samples", type=str, help="path to samples")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="print debug logging. -v extractor info, -vv extractor debug, -vvv cli debug",
    )
    parser.add_argument("--pretty", action="store_true", help="pretty print json output")
    parser.add_argument(
        "--base64",
        action="store_true",
        help="Include base64 encoded binary data in output (can be large, consider printing to file rather than console)",
    )
    parser.add_argument("--logfile", type=str, help="file to log output")
    parser.add_argument("--include", type=str, help="comma separated extractors to run")
    parser.add_argument("--exclude", type=str, help="comma separated extractors to not run")
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="ignore yara rules and execute all extractors",
    )
    parser.add_argument(
        "--create_venv",
        action="store_true",
        help="Creates venvs for every requirements.txt found (only applies when extractor path is a directory)",
    )
    args = parser.parse_args()
    inc = args.include.split(",") if args.include else []
    exc = args.exclude.split(",") if args.exclude else []

    # set up logging for lib, only show debug with 3+ verbose
    logger_lib = logging.getLogger("maco.lib")
    logger_lib.setLevel(logging.DEBUG if args.verbose > 2 else logging.INFO)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    logger_lib.addHandler(ch)

    # set up logging for extractor
    logger_ex = logging.getLogger("maco.extractor")
    if args.verbose == 0:
        logger_ex.setLevel(logging.WARNING)
    elif args.verbose == 1:
        logger_ex.setLevel(logging.INFO)
    else:
        logger_ex.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        fmt="%(asctime)s, [%(levelname)s] %(module)s.%(funcName)s: %(message)s", datefmt="%Y-%m-%d (%H:%M:%S)"
    )
    ch.setFormatter(formatter)
    logger_ex.addHandler(ch)

    # log everything to file
    if args.logfile:
        logger = logging.getLogger("maco")
        logger_lib.setLevel(logging.DEBUG)
        fh = logging.FileHandler(args.logfile)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    process_filesystem(
        args.extractors,
        args.samples,
        inc,
        exc,
        pretty=args.pretty,
        force=args.force,
        include_base64=args.base64,
        create_venv=args.create_venv
    )


if __name__ == "__main__":
    main()



import io
import os
import unittest

from maco import collector

path_extractors = "../../demo_extractors"


class TestHelpersFindExtractors(unittest.TestCase):
    def test_find_extractors(self):
        target = os.path.join(__file__, path_extractors)
        m = collector.Collector(target)
        # extractors = helpers.find_extractors(target)
        self.assertEqual(len(m.extractors), 4)
        self.assertEqual(
            {x for x in m.extractors.keys()},
            {"Complex", "Elfy", "LimitOther", "Nothing"},
        )


class TestHelpersCompileYara(unittest.TestCase):
    def test_compile_yara(self):
        target = os.path.join(__file__, path_extractors)
        m = collector.Collector(target)
        self.assertEqual(
            {x.identifier for x in m.rules},
            {
                "Elfy",
                "Complex",
                "ComplexSubtext",
                "Nothing",
                "ComplexAlt",
                "LimitOther",
            },
        )


class TestHelpersAnalyseStream(unittest.TestCase):
    def setUp(self):
        target = os.path.join(__file__, path_extractors)
        self.m = collector.Collector(target)

    def test_analyse_stream(self):
        data = b""
        resp = self.m.extract(io.BytesIO(data), [], "Complex")
        self.assertEqual(resp, None)

        data = b"data"
        resp = self.m.extract(io.BytesIO(data), [], "Complex")
        self.assertEqual(
            resp,
            {
                "family": "complex",
                "version": "5",
                "binaries": [
                    {
                        "datatype": "payload",
                        "data": b"some data",
                        "encryption": {"algorithm": "something"},
                    }
                ],
                "http": [
                    {
                        "protocol": "https",
                        "hostname": "blarg5.com",
                        "path": "/malz/4",
                        "usage": "c2",
                    }
                ],
                "encryption": [{"algorithm": "sha256"}],
            },
        )



"""Convenience functions for discovering your extractors."""

import logging
import os

from tempfile import NamedTemporaryFile
from typing import Any, BinaryIO, Dict, List

import yara
from pydantic import BaseModel

from maco import extractor, model, utils


class ExtractorLoadError(Exception):
    pass


logger = logging.getLogger("maco.lib.helpers")


def _verify_response(resp: BaseModel) -> Dict:
    """Enforce types and verify properties, and remove defaults."""
    # check the response is valid for its own model
    # this is useful if a restriction on the 'other' dictionary is needed
    resp_model = type(resp)
    if resp_model != model.ExtractorModel:
        resp = resp_model.model_validate(resp)
    # check the response is valid according to the ExtractorModel
    resp = model.ExtractorModel.model_validate(resp)
    # coerce sets to correct types
    # otherwise we end up with sets where we expect lists
    resp = model.ExtractorModel(**resp.model_dump())
    # dump model to dict
    return resp.model_dump(exclude_defaults=True)


class Collector:
    def __init__(
        self, path_extractors: str, include: List[str] = None, exclude: List[str] = None, create_venv: bool = False
    ):
        """Discover and load extractors from file system."""
        self.path = path_extractors
        self.extractors = {}
        namespaced_rules = {}

        if create_venv and os.path.isdir(path_extractors):
            # Recursively create/update virtual environments
            utils.create_venv(path_extractors, logger=logger)

        def extractor_module_callback(member, module, venv) -> bool:
            name = member.__name__
            if exclude and name in exclude:
                # Module is part of the exclusion list, skip
                logger.debug(f"exclude excluded '{name}'")
                return

            if include and name not in include:
                # Module wasn't part of the inclusion list, skip
                logger.debug(f"include excluded '{name}'")
                return

            if utils.maco_extractor_validation(member):
                # check if we want this extractor

                # initialise and register
                logger.debug(f"register '{name}'")
                self.extractors[name] = dict(module=member, venv=venv, module_path=module.__file__)
                namespaced_rules[name] = member.yara_rule or extractor.DEFAULT_YARA_RULE.format(name=name)
                return True

        # Find the extractors within the given directory
        utils.find_extractors(path_extractors, logger=logger, extractor_module_callback=extractor_module_callback)

        if not self.extractors:
            raise ExtractorLoadError("no extractors were loaded")
        logger.debug(f"found extractors {list(self.extractors.keys())}\n")

        # compile yara rules gathered from extractors
        self.rules = yara.compile(sources=namespaced_rules)

    def match(self, stream: BinaryIO) -> Dict[str, List[yara.Match]]:
        """Return extractors that should run based on yara rules."""
        # execute yara rules on file to find extractors we should run
        # yara can't run on a stream so we give it a bytestring
        matches = self.rules.match(data=stream.read())
        stream.seek(0)
        if not matches:
            return
        # get all rules that hit for each extractor
        runs = {}
        for match in matches:
            runs.setdefault(match.namespace, []).append(match)

        return runs

    def extract(
        self,
        stream: BinaryIO,
        matches: List[yara.Match],
        extractor_name: str,
    ) -> Dict[str, Any]:
        """Run extractor with stream and verify output matches the model."""
        extractor = self.extractors[extractor_name]
        resp = None
        try:
            if extractor["venv"]:
                # Run extractor within a virtual environment
                with NamedTemporaryFile() as sample_path:
                    sample_path.write(stream.read())
                    sample_path.flush()
                    return utils.run_in_venv(sample_path.name, **extractor)
            else:
                # Run extractor within on host environment
                resp = extractor["module"]().run(stream, matches)
        except Exception:
            # caller can deal with the exception
            raise
        finally:
            # make sure to reset where we are in the file
            # otherwise follow on extractors are going to read 0 bytes
            stream.seek(0)

        # enforce types and verify properties, and remove defaults
        if resp is not None:
            resp = _verify_response(resp)

        return resp



#!/usr/bin/env python3
"""Setup script."""

from setuptools import setup

setup(
    python_requires=">=3.8",
    use_scm_version={"root": ".."},
    setup_requires=["setuptools_scm"],
    packages=["maco.model"],
    include_package_data=True,
    install_requires=["pydantic>=2.0.0"],
    name="maco-model",
    description="",
    author="",
    author_email="",
    classifiers=[],
)



"""Foundation for unit testing an extractor.

Example:

from maco import base_test
class TestExample(base_test.BaseTest):
    name = "Example"
    path = os.path.join(__file__, "../../extractors")
    def test_run(self):
        data = b"data with Example information"
        ret = self.extract(io.BytesIO(data))
        self.assertEqual(ret["family"], "example")
"""

import importlib
import io
import os
import unittest

import cart

from maco import collector


class NoHitException(Exception):
    pass


class BaseTest(unittest.TestCase):
    name: str = None  # name of the extractor
    # folder and/or file where extractor is.
    # I recommend something like os.path.join(__file__, "../../extractors")
    # if your extractors are in a folder 'extractors' next to a folder of tests
    path: str = None

    def setUp(self) -> None:
        if not self.name or not self.path:
            raise Exception("name and path must be set")
        self.c = collector.Collector(self.path, include=[self.name])
        self.assertIn(self.name, self.c.extractors)
        self.assertEqual(len(self.c.extractors), 1)
        return super().setUp()

    def extract(self, stream):
        """Return results for running extractor over stream, including yara check."""
        runs = self.c.match(stream)
        if not runs:
            raise NoHitException("no yara rule hit")
        hits = runs[self.name]
        resp = self.c.extract(stream, hits, self.name)
        return resp

    def _get_location(self) -> str:
        """Return path to child class that implements this class."""
        # import child module
        module = type(self).__module__
        i = importlib.import_module(module)
        # get location to child module
        return i.__file__

    def load_cart(self, filepath: str) -> io.BytesIO:
        """Load and unneuter a test file (likely malware) into memory for processing."""
        # it is nice if we can load files relative to whatever is implementing base_test
        dirpath = os.path.split(self._get_location())[0]
        # either filepath is absolute, or should be loaded relative to child of base_test
        filepath = os.path.join(dirpath, filepath)
        if not os.path.isfile(filepath):
            raise FileNotFoundError(filepath)
        with open(filepath, "rb") as f:
            unpacked = io.BytesIO()
            # just bubble exceptions if it isn't cart
            cart.unpack_stream(f, unpacked)
        # seek to start of the unneutered stream
        unpacked.seek(0)
        return unpacked



from io import BytesIO
from typing import List, Optional

import yara

from maco import extractor, model


class Bob(extractor.Extractor):
    """A simplistic script for testing."""

    family = "bob"
    author = "bob"
    last_modified = "2022-06-14"



from io import BytesIO
from typing import Dict, List, Optional

import yara

from maco import extractor, model

from . import complex_utils


class Complex(extractor.Extractor):
    """This script has multiple yara rules and coverage of the data model."""

    family = "complex"
    author = "blue"
    last_modified = "2022-06-14"
    yara_rule = """
        private rule ComplexSubtext
        {
            strings:
                $self_trigger = "self_trigger"
            condition:
                $self_trigger
        }
        rule Complex
        {
            strings:
                $self_trigger = "Complex"
                $my_hex_string = { E2 34 A1 C8 23 FB }

            condition:
                ($self_trigger or $my_hex_string) and ComplexSubtext
        }
        rule ComplexAlt
        {
            strings:
                $self_trigger = "Paradise"

            condition:
                $self_trigger
        }
        """

    def run(
        self, stream: BytesIO, matches: List[yara.Match]
    ) -> Optional[model.ExtractorModel]:
        self.logger.info("starting run")
        self.logger.debug(f"{[x.rule for x in matches]=}")
        data = stream.read()
        if not data:
            return
        # this is where you would do some processing on the file
        data_len = len(data)
        other = complex_utils.getdata()["result"]
        self.logger.debug("got data from lib")
        # example - accessing yara strings
        strings = {y[2].decode("utf8") for x in matches for y in x.strings}
        self.logger.debug(f"{strings=}")
        # construct model of results
        tmp = model.ExtractorModel(family=self.family)
        tmp.decoded_strings = strings
        tmp.version = "5"
        tmp.http.append(
            tmp.Http(
                protocol="https",
                hostname=f"blarg{other}.com",
                path=f"/malz/{data_len}",
                usage="c2",
            )
        )

        tmp.encryption.append(tmp.Encryption(algorithm="sha256"))
        tmp.binaries.append(
            tmp.Binary(
                data=b"some data",
                datatype=tmp.Binary.TypeEnum.payload,
                encryption=tmp.Binary.Encryption(algorithm="something"),
            )
        )
        return tmp



from io import BytesIO
from typing import Dict, List, Optional

import yara

from maco import extractor, model


class Nothing(extractor.Extractor):
    """Returns no extracted data."""

    family = "nothing"
    author = "blue"
    last_modified = "2022-06-14"
    yara_rule = """
        rule Nothing
        {
            strings:
                $self_trigger = "Nothing"

            condition:
                $self_trigger
        }
        """

    def run(
        self, stream: BytesIO, matches: List[yara.Match]
    ) -> Optional[model.ExtractorModel]:
        # return config model formatted results
        return



from typing import Optional
import pydantic

from maco import model


class MyCustomModel(model.ExtractorModel):
    class Other(pydantic.BaseModel):
        key1: str
        key2: bool
        key3: int

    # set a custom class here as valid for the 'other' property
    other: Optional[Other] = None



"""Base class for an extractor script."""

import logging
import textwrap
from typing import BinaryIO, List, Optional, Union

import yara

from . import model


class InvalidExtractor(ValueError):
    pass

DEFAULT_YARA_RULE = \
"""
rule {name}
{{
    condition:
        true
}}
"""

class Extractor:
    """Base class for an analysis extractor with common entrypoint and metadata.

    Override this docstring with a good description of your extractor.
    """

    family: Union[str, List[str]] = None  # family or families of malware that is detected by the extractor
    author: str = None  # author of the extractor (name@organisation)
    last_modified: str = None  # last modified date (YYYY-MM-DD)
    sharing: str = "TLP:WHITE"  # who can this be shared with?
    yara_rule: str = None  # yara rule that we filter inputs with
    reference: str = None  # link to malware report or other reference information
    logger: logging.Logger = None  # logger for use when debugging

    def __init__(self) -> None:
        self.name = name = type(self).__name__
        self.logger = logging.getLogger(f"maco.extractor.{name}")
        self.logger.debug(f"initialise '{name}'")
        if not self.family or not self.author or not self.last_modified:
            raise InvalidExtractor("must set family, author, last_modified")
        # if author does not set a yara rule, match on everything
        if not self.yara_rule:
            self.yara_rule = DEFAULT_YARA_RULE.format(name=name)
        # unindent the yara rule from triple quoted string
        # this is for friendly printing, yara handles the rule ok either way
        self.yara_rule = textwrap.dedent(self.yara_rule)
        # check yara rules conform to expected structure
        # we throw away these compiled rules as we need all rules in system compiled together
        try:
            rules = yara.compile(source=self.yara_rule)
        except yara.SyntaxError as e:
            raise InvalidExtractor(f"{self.name} - invalid yara rule") from e
        # need to track which plugin owns the rules
        self.yara_rule_names = [x.identifier for x in rules]
        if not len(list(rules)):
            raise InvalidExtractor(f"{name} must define at least one yara rule")
        for x in rules:
            if x.is_global:
                raise InvalidExtractor(f"{x.identifier} yara rule must not be global")

    def run(self, stream: BinaryIO, matches: List[yara.Match]) -> Optional[model.ExtractorModel]:
        """Run the analysis process and return dict matching.

        :param stream: file object from disk/network/memory.
        :param match: yara rule match information contains locations of strings.
        """
        raise NotImplementedError()



from io import BytesIO
from typing import Dict, List, Optional

import yara

from maco import extractor, model


class Elfy(extractor.Extractor):
    """Check basic elf property."""

    family = "elfy"
    author = "blue"
    last_modified = "2022-06-14"
    yara_rule = """
        import "elf"

        rule Elfy
        {
            condition:
                elf.number_of_sections > 50
        }
        """

    def run(
        self, stream: BytesIO, matches: List[yara.Match]
    ) -> Optional[model.ExtractorModel]:
        # return config model formatted results
        ret = model.ExtractorModel(family=self.family)
        # the list for campaign_id already exists and is empty, so we just add an item
        ret.campaign_id.append(str(len(stream.read())))
        return ret



import os

from maco import base_test


class TestLimitOther(base_test.BaseTest):
    name = "LimitOther"
    path = os.path.join(__file__, "../../demo_extractors")

    def test_load_cart(self):
        data = self.load_cart("data/example.txt.cart").read()
        self.assertEqual(data, b"LimitOther\n")

    def test_run(self):
        ret = self.extract(self.load_cart("data/example.txt.cart"))
        self.assertEqual(ret["family"], "specify_other")
        self.assertEqual(ret["campaign_id"], ["12345"])



from io import BytesIO
from typing import Dict, List, Optional

import yara

from maco import extractor, model

from . import shared


class LimitOther(extractor.Extractor):
    """An example of how the 'other' dictionary can be limited in a custom way."""

    family = "limit_other"
    author = "blue"
    last_modified = "2022-06-14"
    yara_rule = """
        rule LimitOther
        {
            strings:
                $self_trigger = "LimitOther"

            condition:
                $self_trigger
        }
        """

    def run(
        self, stream: BytesIO, matches: List[yara.Match]
    ) -> Optional[model.ExtractorModel]:
        # use a custom model that inherits from ExtractorModel
        # this model defines what can go in the 'other' dict
        tmp = shared.MyCustomModel(family="specify_other")
        tmp.campaign_id.append("12345")
        tmp.other = tmp.Other(key1="key1", key2=True, key3=45)
        return tmp



"""Malware config extractor output model."""

from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict


class ForbidModel(BaseModel):
    """We want to forbid extra properties, so that the 'other' field is used instead."""

    model_config = ConfigDict(extra="forbid", use_enum_values=True)


class ConnUsageEnum(str, Enum):
    """Purpose of the connection."""

    c2 = "c2"  # issue commands to malware
    upload = "upload"  # get data out of the network
    download = "download"  # fetch dynamic config, second stage, etc
    propagate = "propagate"  # spread through the network
    tunnel = "tunnel"  # communicate through the network
    ransom = "ransom"  # payment
    decoy = "decoy"  # Decoy connections to obfuscate malicious
    other = "other"


class Encryption(ForbidModel):
    """Encryption usage."""

    class UsageEnum(str, Enum):
        config = "config"
        communication = "communication"
        binary = "binary"
        ransom = "ransom"
        other = "other"

    algorithm: Optional[str] = None
    public_key: Optional[str] = None
    key: Optional[str] = None  # private key or symmetric key
    provider: Optional[str] = None  # encryption library used. openssl, homebrew, etc.

    mode: Optional[str] = None  # block vs stream
    # base 64'd binary data for these details?
    # TODO to confirm usage of these different properties
    iv: Optional[str] = None  # initialisation vector
    seed: Optional[str] = None
    nonce: Optional[str] = None
    constants: List[str] = []

    usage: Optional[UsageEnum] = None


class CategoryEnum(str, Enum):
    # Software that shows you extra promotions that you cannot control as you use your PC.
    # You wouldn't see the extra ads if you didn't have adware installed.
    adware = "adware"

    # Malware related to an Advanced Persistent Threat (APT) group.
    apt = "apt"

    # A backdoor Trojan gives malicious users remote control over the infected computer. They enable the author to do anything they wish on the infected computer including sending, receiving, launching and deleting files, displaying data and rebooting the computer. Backdoor Trojans are often used to unite a group of victim computers to form a botnet or zombie network that can be used for criminal purposes.
    backdoor = "backdoor"

    # Trojan Banker programs are designed to steal your account data for online banking systems, e-payment systems and credit or debit cards.
    banker = "banker"

    # A malware variant that modifies the boot sectors of a hard drive, including the Master Boot Record (MBR) and Volume Boot Record (VBR).
    bootkit = "bootkit"

    # A malicious bot is self-propagating malware designed to infect a host and connect back to a central server or servers that act as a command and control (C&C) center for an entire network of compromised devices, or botnet.
    bot = "bot"

    # A browser hijacker is defined as a form of unwanted software that modifies a web browser's settings without the user's permission. The result is the placement of unwanted advertising into the browser, and possibly the replacement of an existing home page or search page with the hijacker page.
    browser_hijacker = "browser_hijacker"

    # Trojan bruteforcer are trying to brute force website in order to achieve something else (EX: Finding  WordPress websites with default credentials).
    bruteforcer = "bruteforcer"

    # A type of trojan that can use your PC to 'click' on websites or applications. They are usually used to make money for a malicious hacker by clicking on online advertisements and making it look like the website gets more traffic than it does. They can also be used to skew online polls, install programs on your PC, or make unwanted software appear more popular than it is.
    clickfraud = "clickfraud"

    # Cryptocurrency mining malware.
    cryptominer = "cryptominer"

    # These programs conduct DoS (Denial of Service) attacks against a targeted web address. By sending multiple requests from your computer and several other infected computers, the attack can overwhelm the target address leading to a denial of service.
    ddos = "ddos"

    # Trojan Downloaders can download and install new versions of malicious programs in the target system.
    downloader = "downloader"

    # These programs are used by hackers in order to install malware or to prevent the detection of malicious programs.
    dropper = "dropper"

    # Exploit kits are programs that contain data or code that takes advantage of a vulnerability within an application that is running in the target system.
    exploitkit = "exploitkit"

    # Trojan FakeAV programs simulate the activity of antivirus software. They are designed to extort money in return for the detection and removal of threat, even though the threats that they report are actually non-existent.
    fakeav = "fakeav"

    # A type of tool that can be used to allow and maintain unauthorized access to your PC.
    hacktool = "hacktool"

    # A program that collects your personal information, such as your browsing history, and uses it without adequate consent.
    infostealer = "infostealer"

    # A keylogger monitors and logs every keystroke it can identify. Once installed, the virus either keeps track of all the keys and stores the information locally, after which the hacker needs physical access to the computer to retrieve the information, or the logs are sent over the internet back to the hacker.
    keylogger = "keylogger"

    # A program that loads another application / memory space.
    loader = "loader"

    # A type of malware that hides its code and purpose to make it more difficult for security software to detect or remove it.
    obfuscator = "obfuscator"

    # Point-of-sale malware is usually a type of malware that is used by cybercriminals to target point of sale (POS) and payment terminals with the intent to obtain credit card and debit card information.
    pos = "pos"

    # This type of trojan allows unauthorized parties to use the infected computer as a proxy server to access the Internet anonymously.
    proxy = "proxy"

    # A program that can be used by a remote hacker to gain access and control of an infected machine.
    rat = "rat"

    # This type of malware can modify data in the target computer so the operating system will stop running correctly or the data is no longer accessible. The criminal will only restore the computer state or data after a ransom is paid to them (mostly using cryptocurrency).
    ransomware = "ransomware"

    # A reverse proxy is a server that receives requests from the internet and forwards them to a small set of servers.
    reverse_proxy = "reverse_proxy"

    # Rootkits are designed to conceal certain objects or activities in the system. Often their main purpose is to prevent malicious programs being detected in order to extend the period in which programs can run on an infected computer.
    rootkit = "rootkit"

    # This type of malware scan the internet / network(s) / system(s) / service(s) to collect information. That information could be used later to perpetuate an cyber attack.
    scanner = "scanner"

    # Scareware is a form of malware which uses social engineering to cause shock, anxiety, or the perception of a threat in order to manipulate users into buying unwanted software.
    scareware = "scareware"

    # Malware that is sending spam.
    spammer = "spammer"

    # Generic or Unknown Trojan
    trojan = "trojan"

    # A generic computer virus
    virus = "virus"

    # A type of malware that destroy the data.
    wiper = "wiper"

    # A web shell is a script that can be uploaded to a web server to enable remote administration of the machine.
    webshell = "webshell"

    # A type of malware that spreads to other PCs.
    worm = "worm"


class ExtractorModel(ForbidModel):
    r"""Captured config/iocs, unpacked binaries and other malware properties from a robo-analyst.

    This model defines common fields for output of a script targeting a specific malware family.
    Usage of this model will allow for easier sharing of scripts between different authors and systems.
    The model will not define fields for all data that can be extracted from a binary, only the most common.
    This is to make it easier for authors to understand and use the model.

    This model can have new fields added in the future if they become more common,
    but the intent is to avoid removing or modifying existing fields, for backwards compatibility.

    Where data does not fit with the current model, the 'others' field should be used.
    Contents in this field is not defined by the model and verification/normalisation is up to
    the author and whatever systems run the scripts.
    If many decoders define similar data in the 'others' field, that field should be migrated to this model.

    The model must be kept relatively flat, with nested lists of dictionaries to be avoided.
    This is to make queries simpler to write in sql, elasticsearch and other storage systems.

    Malware and systems that investigate malware can do pretty much anything.
    This model needs to be simple and flexible to make sharing easy.
    Some things should be out of scope for this model.
    Responsibility for these things are up to authors and systems that use this model.

    Out of scope
    * Verifying anything in the 'others' dict, including that it is json-compatible.
        * We don't know anything about the structure
        * checking is json compatible requires dumping to json string, which can be slow
    * Connecting specific config items to malware behaviour catalog
        * i.e. "Persistence::Modify Registry" with 'registry' item from model (SYSTEM\ControlSet001\Services\)
        * due to complexity and normalisation difficulties
        * much malware behaviour is not related to specific config items
    * Normalisation/verification of individual properties
        * i.e. lowercase filepaths - some filesystems are case sensitive
        * i.e. checking registry hives match known - not enough SME and too complex for a simple model
        * generally, this quickly becomes complex (validating a fully defined http item)
        * calling systems are probably performing their own validation anyway
    * requiring specific properties to be set
        * i.e. if http item is defined, requiring hostname to be set
        * Some use cases always seem to exist where a property should not be set
    """

    family: Union[str, List[str]]  # family or families of malware that was detected
    version: Optional[str] = None  # version/variant of malware
    category: List[CategoryEnum] = []  # capability/purpose of the malware
    attack: List[str] = []  # mitre att&ck reference ids, e.g. 'T1129'

    #
    # simple config properties
    #

    # capabilities of the malware enabled/disabled in config
    # note these are probably malware-specific capabilities so no attempt to normalise has been made
    # note - av/sandbox detection should be noted by 'detect_<product>'
    capability_enabled: List[str] = []
    capability_disabled: List[str] = []

    campaign_id: List[str] = []  # Server/Campaign Id for malware
    identifier: List[str] = []  # UUID/Identifiers for deployed instance
    decoded_strings: List[str] = []  # decoded strings from within malware
    password: List[str] = []  # Any password extracted from the binary
    mutex: List[str] = []  # mutex to prevent multiple instances
    pipe: List[str] = []  # pipe name used for communication
    sleep_delay: Optional[int] = None  # time to sleep/delay execution (milliseconds)
    # additional time applied to sleep_delay (milliseconds).
    # Jitter implementations can vary but usually it is a value from which a random number is generated and
    # added/subtracted to the sleep_delay to make behaviour more unpredictable
    sleep_delay_jitter: Optional[int] = None
    inject_exe: List[str] = []  # name of executable to inject into

    # configuration or clustering/research data that doesnt fit the other fields
    # * rarely used by decoders or specific to one decoder
    # to prevent key explosion, the keys must not be dynamically generated
    # e.g. api_imports, api_checksums, num_imports, import_hash + many more
    # data stored here must always be JSON-serialisable
    other: Dict[str, Any] = {}

    #
    # embedded binary data
    #
    class Binary(ForbidModel):
        """Binary data extracted by decoder."""

        class TypeEnum(str, Enum):
            payload = "payload"  # contained within the original file
            config = "config"  # sometimes malware uses json/formatted text for config
            other = "other"

        datatype: Optional[TypeEnum] = None  # what the binary data is used for
        data: bytes  # binary data, not json compatible

        # other information for the extracted binary rather than the config
        # data stored here must always be JSON-serialisable
        # e.g. filename, extension, relationship label
        other: Dict[str, Any] = {}

        # convenience for ret.encryption.append(ret.Encryption(*properties))
        # Define as class as only way to allow for this to be accessed and not have pydantic try to parse it.
        class Encryption(Encryption):
            pass

        encryption: Union[List[Encryption], Encryption, None] = None  # encryption information for the binary

    binaries: List[Binary] = []

    #
    # communication protocols
    #
    class FTP(ForbidModel):
        """Usage of FTP connection."""

        username: Optional[str] = None
        password: Optional[str] = None
        hostname: Optional[str] = None
        port: Optional[int] = None

        path: Optional[str] = None

        usage: Optional[ConnUsageEnum] = None

    ftp: List[FTP] = []

    class SMTP(ForbidModel):
        """Usage of SMTP."""

        # credentials and location of server
        username: Optional[str] = None
        password: Optional[str] = None
        hostname: Optional[str] = None
        port: Optional[int] = None

        mail_to: List[str] = []  # receivers
        mail_from: Optional[str] = None  # sender
        subject: Optional[str] = None

        usage: Optional[ConnUsageEnum] = None

    smtp: List[SMTP] = []  # SMTP server for malware

    class Http(ForbidModel):
        """Usage of HTTP connection."""

        # malware sometimes does weird stuff with uris so we don't want to force
        # authors to break the uri into username, hostname, path, etc.
        # as we lose that information.
        # e.g. extra '?' or '/' when unnecessary.
        # or something that is technically an invalid uri but still works
        uri: Optional[str] = None

        # on the other hand we might not have enough info to construct a uri
        protocol: Optional[str] = None  # http,https
        username: Optional[str] = None
        password: Optional[str] = None
        hostname: Optional[str] = None  # (A host/hostname can be an IP, domain or hostname)
        port: Optional[int] = None
        path: Optional[str] = None
        query: Optional[str] = None
        fragment: Optional[str] = None

        user_agent: Optional[str] = None  # user agent sent by malware
        method: Optional[str] = None  # get put delete etc
        headers: Optional[Dict[str, str]] = None  # custom/additional HTTP headers
        max_size: Optional[int] = None

        usage: Optional[ConnUsageEnum] = None

    http: List[Http] = []

    class SSH(ForbidModel):
        """Usage of ssh connection."""

        username: Optional[str] = None
        password: Optional[str] = None
        hostname: Optional[str] = None
        port: Optional[int] = None

        usage: Optional[ConnUsageEnum] = None

    ssh: List[SSH] = []

    class Proxy(ForbidModel):
        """Usage of proxy connection."""

        protocol: Optional[str] = None  # socks5,http
        username: Optional[str] = None
        password: Optional[str] = None
        hostname: Optional[str] = None
        port: Optional[int] = None

        usage: Optional[ConnUsageEnum] = None

    proxy: List[Proxy] = []

    #
    # inter process communication (IPC)
    #
    class IPC(ForbidModel):
        """Usage of named pipe communications."""

        # A record stored on disk, or a record synthesized on demand by a file
        # server, which can be accessed by multiple processes.
        file: Optional[List[str]] = None
        # Data sent over a network interface, either to a different process on
        # the same computer or to another computer on the network. Stream
        # oriented (TCP; data written through a socket requires formatting to
        # preserve message boundaries) or more rarely message-oriented (UDP,
        # SCTP).
        socket: Optional[List[str]] = None
        # Similar to an internet socket, but all communication occurs within
        # the kernel. Domain sockets use the file system as their address
        # space. Processes reference a domain socket as an inode, and multiple
        # processes can communicate with one socket.
        unix_domain_socket: Optional[List[str]] = None
        # A file mapped to RAM and can be modified by changing memory
        # addresses directly instead of outputting to a stream. This shares
        # the same benefits as a standard file.
        memory_mapped_file: Optional[Union[bytes, List[str]]] = None
        # A data stream similar to a socket, but which usually preserves
        # message boundaries. Typically implemented by the operating system,
        # they allow multiple processes to read and write to the message queue
        # without being directly connected to each other.
        message_queue: Optional[List[str]] = None
        # A unidirectional data channel using standard input and output. Data
        # written to the write-end of the pipe is buffered by the operating
        # system until it is read from the read-end of the pipe. Two-way
        # communication between processes can be achieved by using two pipes
        # in opposite "directions".
        anonymous_pipe: Optional[List[str]] = None
        # A pipe that is treated like a file. Instead of using standard input
        # and output as with an anonymous pipe, processes write to and read
        # from a named pipe, as if it were a regular file.
        named_pipe: Optional[List[str]] = None
        # The process names involved in the IPC communication
        process_names: Optional[List[str]] = None
        # Multiple processes are given access to the same block of memory,
        # which creates a shared buffer for the processes to communicate with
        # each other.
        shared_memory: Optional[bytes] = None
        usage: Optional[ConnUsageEnum] = None

    ipc: List[IPC] = []  # Inter-Process Communications (similar to 'pipe' but more detailed)

    class DNS(ForbidModel):
        """Direct usage of DNS."""

        class RecordTypeEnum(str, Enum):
            A = "A"
            AAAA = "AAAA"
            AFSDB = "AFSDB"
            APL = "APL"
            CAA = "CAA"
            CDNSKEY = "CDNSKEY"
            CDS = "CDS"
            CERT = "CERT"
            CNAME = "CNAME"
            CSYNC = "CSYNC"
            DHCID = "DHCID"
            DLV = "DLV"
            DNAME = "DNAME"
            DNSKEY = "DNSKEY"
            DS = "DS"
            EUI48 = "EUI48"
            EUI64 = "EUI64"
            HINFO = "HINFO"
            HIP = "HIP"
            HTTPS = "HTTPS"
            IPSECKEY = "IPSECKEY"
            KEY = "KEY"
            KX = "KX"
            LOC = "LOC"
            MX = "MX"
            NAPTR = "NAPTR"
            NS = "NS"
            NSEC = "NSEC"
            NSEC3 = "NSEC3"
            NSEC3PARAM = "NSEC3PARAM"
            OPENPGPKEY = "OPENPGPKEY"
            PTR = "PTR"
            RRSIG = "RRSIG"
            RP = "RP"
            SIG = "SIG"
            SMIMEA = "SMIMEA"
            SOA = "SOA"
            SRV = "SRV"
            SSHFP = "SSHFP"
            SVCB = "SVCB"
            TA = "TA"
            TKEY = "TKEY"
            TLSA = "TLSA"
            TSIG = "TSIG"
            TXT = "TXT"
            URI = "URI"
            ZONEMD = "ZONEMD"

        ip: Optional[str] = None
        port: Optional[int] = None  # The default value is 53
        hostname: Optional[str] = None  # This is the query hostname
        record_type: Optional[RecordTypeEnum] = None  # The DNS record type that is queried
        usage: Optional[ConnUsageEnum] = None

    dns: List[DNS] = []  # custom DNS address to use for name resolution

    class Connection(ForbidModel):
        """Generic TCP/UDP usage."""

        client_ip: Optional[str] = None
        client_port: Optional[int] = None
        server_ip: Optional[str] = None
        server_domain: Optional[str] = None
        server_port: Optional[int] = None

        usage: Optional[ConnUsageEnum] = None

    tcp: List[Connection] = []
    udp: List[Connection] = []

    #
    # complex configuration properties
    #
    # convenience for ret.encryption.append(ret.Encryption(*properties))
    # Define as class as only way to allow for this to be accessed and not have pydantic try to parse it.
    class Encryption(Encryption):
        pass

    encryption: List[Encryption] = []

    class Service(ForbidModel):
        """OS service usage by malware."""

        dll: Optional[str] = None  # dll that the service is loaded from
        name: Optional[str] = None  # service/driver name for persistence
        display_name: Optional[str] = None  # display name for service
        description: Optional[str] = None  # description for service

    service: List[Service] = []

    class Cryptocurrency(ForbidModel):
        """Cryptocoin usage (ransomware/miner)."""

        class UsageEnum(str, Enum):
            ransomware = "ransomware"  # request money to unlock
            miner = "miner"  # use gpu/cpu to mint coins
            other = "other"

        coin: Optional[str] = None  # BTC,ETH,USDT,BNB, etc
        address: Optional[str] = None
        ransom_amount: Optional[float] = None  # number of coins required (if hardcoded)

        usage: UsageEnum

    cryptocurrency: List[Cryptocurrency] = []

    class Path(ForbidModel):
        class UsageEnum(str, Enum):
            c2 = "c2"  # file/folder issues commands to malware
            config = "config"  # config is loaded from this path
            install = "install"  # install directory/filename for malware
            plugins = "plugins"  # load new capability from this directory
            logs = "logs"  # location to log activity
            storage = "storage"  # location to store/backup copied files
            other = "other"

        # C:\User\tmp\whatever.txt or /some/unix/folder/path
        path: str
        usage: Optional[UsageEnum] = None

    paths: List[Path] = []  # files/directories used by malware

    class Registry(ForbidModel):
        class UsageEnum(str, Enum):
            persistence = "persistence"  # stay alive
            store_data = "store_data"  # generated encryption keys or config
            store_payload = "store_payload"  # malware hidden in registry key
            read = "read"  # read system registry keys
            other = "other"

        key: str
        usage: Optional[UsageEnum] = None

    registry: List[Registry] = []



from io import BytesIO
from typing import List, Optional

import yara

from maco import extractor, model


class Basic(extractor.Extractor):
    """A simplistic script for testing."""

    family = "basic"
    author = "blue"
    last_modified = "2022-06-14"
    yara_rule = """
        rule Basic
        {
            strings:
                $self_trigger = "Basic"

            condition:
                $self_trigger
        }
        """

    def run(self, stream: BytesIO, matches: List[yara.Match]) -> Optional[model.ExtractorModel]:
        # use a custom model that inherits from ExtractorModel
        # this model defines what can go in the 'other' dict
        tmp = model.ExtractorModel(family="basic")
        tmp.campaign_id.append("12345")
        tmp.other = dict(key1="key1", key2=True, key3=45)
        return tmp



import unittest
from typing import Dict

from pydantic import ValidationError

from maco import collector, model


class TestModelObject(unittest.TestCase):
    maxDiff = None

    def test_model_invalid(self):
        # family not supplied
        self.assertRaises(ValidationError, model.ExtractorModel)

        ret = model.ExtractorModel(family="octopus")
        # invalid property
        self.assertRaises(ValueError, setattr, *(ret, "invalid", 12345))
        # invalid type
        ret.sleep_delay = "test"
        self.assertRaises(ValidationError, collector._verify_response, ret)

    def test_model_object_1(self):
        # object example
        tmp = model.ExtractorModel(family="scuba")
        tmp.campaign_id.append("5467")
        self.verify(tmp, {"family": "scuba", "campaign_id": ["5467"]})

    def test_model_object_2(self):
        em = model.ExtractorModel
        tmp = model.ExtractorModel(
            family="scuba",
            version="lotso_stuff",
            category=[],
            attack=[],
            capability_enabled=[],
            capability_disabled=[],
            campaign_id=["32"],
            identifier=["uxuduxuduxuudux"],
            decoded_strings=["there", "are", "some", "strings"],
            password=["hunter2"],
            mutex=["YEAH"],
            pipe=["xiod"],
            sleep_delay=45000,
            sleep_delay_jitter=2500,
            inject_exe=["Teams.exe"],
            other={"misc_data": {"nested": 5}},
            binaries=[
                em.Binary(
                    datatype=None,
                    data=b"\x10\x20\x30\x40",
                    other={
                        "datatype": ["payload"],
                        "extension": [".invalid"],
                        "label": ["xor 0x04 at 0x2130-0x2134"],
                        "some_junk": [1, 2, 3, 4, 5, 6],
                    },
                    encryption=em.Binary.Encryption(
                        algorithm="alxor",
                        public_key=None,
                        key=None,
                        provider=None,
                        mode=None,
                        iv=None,
                        seed=None,
                        nonce=None,
                        constants=[],
                        usage="binary",
                    ),
                ),
                em.Binary(
                    datatype=None,
                    data=b"\x50\x60\x70\x80",
                    other={"datatype": ["payload"]},
                    encryption=[
                        em.Binary.Encryption(
                            algorithm="alxor",
                            public_key=None,
                            key=None,
                            provider=None,
                            mode=None,
                            iv=None,
                            seed=None,
                            nonce=None,
                            constants=[],
                            usage="binary",
                        ),
                        em.Binary.Encryption(
                            algorithm="RC4",
                            public_key=None,
                            key=None,
                            provider=None,
                            mode=None,
                            iv=None,
                            seed=None,
                            nonce=None,
                            constants=[],
                            usage="binary",
                        ),
                    ],
                ),
            ],
            ftp=[
                em.FTP(
                    username=None,
                    password=None,
                    hostname="somewhere",
                    port=None,
                    path=None,
                    usage="c2",
                )
            ],
            smtp=[
                em.SMTP(
                    username=None,
                    password=None,
                    hostname="here.com",
                    port=None,
                    mail_to=[],
                    mail_from=None,
                    subject=None,
                    usage="upload",
                )
            ],
            http=[
                em.Http(
                    uri=None,
                    protocol="https",
                    username=None,
                    password=None,
                    hostname="blarg.com",
                    port=None,
                    path="/malz",
                    query=None,
                    fragment=None,
                    user_agent=None,
                    method=None,
                    headers=None,
                    max_size=None,
                    usage="c2",
                )
            ],
            ssh=[
                em.SSH(
                    username=None,
                    password=None,
                    hostname="bad.malware",
                    port=None,
                    usage="download",
                )
            ],
            proxy=[
                em.Proxy(
                    protocol=None,
                    username=None,
                    password=None,
                    hostname="192.168.0.80",
                    port=None,
                    usage="tunnel",
                )
            ],
            dns=[em.DNS(ip="123.21.21.21", port=None, usage="other")],
            tcp=[
                em.Connection(
                    client_ip=None,
                    client_port=None,
                    server_ip="73.21.32.43",
                    server_domain=None,
                    server_port=None,
                    usage="c2",
                )
            ],
            udp=[
                em.Connection(
                    client_ip=None,
                    client_port=None,
                    server_ip="73.21.32.43",
                    server_domain=None,
                    server_port=None,
                    usage="c2",
                )
            ],
            encryption=[
                em.Encryption(
                    algorithm="alxor",
                    public_key=None,
                    key=None,
                    provider=None,
                    mode=None,
                    iv=None,
                    seed=None,
                    nonce=None,
                    constants=[],
                    usage="binary",
                )
            ],
            service=[
                em.Service(
                    dll=None,
                    name="DeviceMonitorSvc",
                    display_name="DeviceMonitorSvc",
                    description="Device Monitor Service",
                )
            ],
            cryptocurrency=[
                em.Cryptocurrency(
                    coin="APE",
                    address="689fdh658790d6dr987yth84iyth7er8gtrfohyt9",
                    ransom_amount=None,
                    usage="miner",
                )
            ],
            paths=[
                em.Path(path="C:/Windows/system32", usage="install"),
                em.Path(path="C:/user/USERNAME/xxxxx/xxxxx/", usage="logs"),
                em.Path(path="\\here\\is\\some\\place", usage="install"),
            ],
            registry=[
                em.Registry(key="HKLM_LOCAL_USER/some/location/to/key", usage="store_data"),
                em.Registry(key="HKLM_LOCAL_USER/system/location", usage="read"),
            ],
        )
        self.verify(
            tmp,
            {
                "family": "scuba",
                "version": "lotso_stuff",
                "campaign_id": ["32"],
                "identifier": ["uxuduxuduxuudux"],
                "decoded_strings": ["there", "are", "some", "strings"],
                "password": ["hunter2"],
                "mutex": ["YEAH"],
                "pipe": ["xiod"],
                "sleep_delay": 45000,
                "sleep_delay_jitter": 2500,
                "inject_exe": ["Teams.exe"],
                "other": {"misc_data": {"nested": 5}},
                "binaries": [
                    {
                        "data": b"\x10 0@",
                        "other": {
                            "datatype": ["payload"],
                            "extension": [".invalid"],
                            "label": ["xor 0x04 at 0x2130-0x2134"],
                            "some_junk": [1, 2, 3, 4, 5, 6],
                        },
                        "encryption": {"algorithm": "alxor", "usage": "binary"},
                    },
                    {
                        "data": b"P`p\x80",
                        "other": {"datatype": ["payload"]},
                        "encryption": [
                            {"algorithm": "alxor", "usage": "binary"},
                            {"algorithm": "RC4", "usage": "binary"},
                        ],
                    },
                ],
                "ftp": [{"hostname": "somewhere", "usage": "c2"}],
                "smtp": [{"hostname": "here.com", "usage": "upload"}],
                "http": [
                    {
                        "protocol": "https",
                        "hostname": "blarg.com",
                        "path": "/malz",
                        "usage": "c2",
                    }
                ],
                "ssh": [{"hostname": "bad.malware", "usage": "download"}],
                "proxy": [{"hostname": "192.168.0.80", "usage": "tunnel"}],
                "dns": [{"ip": "123.21.21.21", "usage": "other"}],
                "tcp": [{"server_ip": "73.21.32.43", "usage": "c2"}],
                "udp": [{"server_ip": "73.21.32.43", "usage": "c2"}],
                "encryption": [{"algorithm": "alxor", "usage": "binary"}],
                "service": [
                    {
                        "name": "DeviceMonitorSvc",
                        "display_name": "DeviceMonitorSvc",
                        "description": "Device Monitor Service",
                    }
                ],
                "cryptocurrency": [
                    {
                        "coin": "APE",
                        "address": "689fdh658790d6dr987yth84iyth7er8gtrfohyt9",
                        "usage": "miner",
                    }
                ],
                "paths": [
                    {"path": "C:/Windows/system32", "usage": "install"},
                    {"path": "C:/user/USERNAME/xxxxx/xxxxx/", "usage": "logs"},
                    {"path": "\\here\\is\\some\\place", "usage": "install"},
                ],
                "registry": [
                    {
                        "key": "HKLM_LOCAL_USER/some/location/to/key",
                        "usage": "store_data",
                    },
                    {"key": "HKLM_LOCAL_USER/system/location", "usage": "read"},
                ],
            },
        )

    def verify(self, in1, in2: Dict) -> Dict:
        """Verify the returned data matches the schema."""
        resp = collector._verify_response(in1)
        self.assertEqual(resp, in2)


class TestModelDict(unittest.TestCase):
    def test_model_1(self):
        # dict example
        self.verify(
            {
                "family": "scuba",
                "version": "30-01-2023",
                "http": [
                    {
                        "protocol": "https",
                        "hostname": "blarg.com",
                        "path": "/malz",
                        "usage": "c2",
                    }
                ],
            }
        )

    def test_model_2(self):
        # dict example large
        self.maxDiff = None

        self.verify(
            {
                "family": "scuba",
                "version": "lotso_stuff",
                "binaries": [
                    {
                        "data": rb"\x10\x20\x30\x40",
                        "encryption": {"algorithm": "alxor", "usage": "binary"},
                        "other": {
                            "datatype": ["payload"],
                            "extension": [".invalid"],
                            "label": ["xor 0x04 at 0x2130-0x2134"],
                            "some_junk": [1, 2, 3, 4, 5, 6],
                        },
                    },
                    {
                        "data": rb"\x50\x60\x70\x80",
                        "encryption": [
                            {"algorithm": "alxor", "usage": "binary"},
                            {"algorithm": "RC4", "usage": "binary"},
                        ],
                        "other": {
                            "datatype": ["payload"],
                        },
                    },
                ],
                "ftp": [{"hostname": "somewhere", "usage": "c2"}],
                "smtp": [{"hostname": "here.com", "usage": "upload"}],
                "http": [
                    {
                        "protocol": "https",
                        "hostname": "blarg.com",
                        "path": "/malz",
                        "usage": "c2",
                    }
                ],
                "ssh": [{"hostname": "bad.malware", "usage": "download"}],
                "proxy": [{"hostname": "192.168.0.80", "usage": "tunnel"}],
                "dns": [{"ip": "123.21.21.21", "usage": "other"}],
                "tcp": [{"server_ip": "73.21.32.43", "usage": "c2"}],
                "udp": [{"server_ip": "73.21.32.43", "usage": "c2"}],
                "encryption": [{"algorithm": "alxor", "usage": "binary"}],
                "service": [
                    {
                        "name": "DeviceMonitorSvc",
                        "display_name": "DeviceMonitorSvc",
                        "description": "Device Monitor Service",
                    }
                ],
                "cryptocurrency": [
                    {
                        "coin": "APE",
                        "address": "689fdh658790d6dr987yth84iyth7er8gtrfohyt9",
                        "usage": "miner",
                    }
                ],
                "paths": [
                    {"path": "C:/Windows/system32", "usage": "install"},
                    {"path": "C:/user/USERNAME/xxxxx/xxxxx/", "usage": "logs"},
                    {"path": "\\here\\is\\some\\place", "usage": "install"},
                ],
                "registry": [
                    {
                        "key": "HKLM_LOCAL_USER/some/location/to/key",
                        "usage": "store_data",
                    },
                    {"key": "HKLM_LOCAL_USER/system/location", "usage": "read"},
                ],
                "campaign_id": ["32"],
                "identifier": ["uxuduxuduxuudux"],
                "decoded_strings": ["there", "are", "some", "strings"],
                "password": ["hunter2"],
                "mutex": ["YEAH"],
                "pipe": ["xiod"],
                "sleep_delay": 45000,
                "inject_exe": ["Teams.exe"],
                "other": {"misc_data": {"nested": 5}},
            }
        )

    def verify(self, config: Dict) -> Dict:
        """Verify the returned data matches the schema."""
        tmp = model.ExtractorModel.model_validate(config)
        resp = collector._verify_response(tmp)
        self.assertEqual(resp, config)



import os
import pytest
import sys

from maco.collector import Collector

INIT_MODULES = list(sys.modules.keys())

CAPE_EXTRACTORS = [
    "AgentTesla",
    "AsyncRAT",
    "AuroraStealer",
    "Azorult",
    "BackOffLoader",
    "BackOffPOS",
    "BitPaymer",
    "BlackDropper",
    "BlackNix",
    "Blister",
    "BruteRatel",
    "BuerLoader",
    "BumbleBee",
    "Carbanak",
    "ChChes",
    "CobaltStrikeBeacon",
    "CobaltStrikeStager",
    "DCRat",
    "DarkGate",
    "DoppelPaymer",
    "DridexLoader",
    "Emotet",
    "Enfal",
    "EvilGrab",
    "Fareit",
    "Formbook",
    "Greame",
    "GuLoader",
    "Hancitor",
    "HttpBrowser",
    "IcedID",
    "IcedIDLoader",
    "KoiLoader",
    "Latrodectus",
    "LokiBot",
    "Lumma",
    "NanoCore",
    "Nighthawk",
    "Njrat",
    "Oyster",
    "Pandora",
    "PhemedroneStealer",
    "PikaBot",
    "PlugX",
    "PoisonIvy",
    "Punisher",
    "QakBot",
    "QuasarRAT",
    "Quickbind",
    "RCSession",
    "REvil",
    "RedLeaf",
    "RedLine",
    "Remcos",
    "Retefe",
    "Rhadamanthys",
    "Rozena",
    "SmallNet",
    "SmokeLoader",
    "Socks5Systemz",
    "SparkRAT",
    "SquirrelWaffle",
    "Stealc",
    "Strrat",
    "TSCookie",
    "TrickBot",
    "UrsnifV3",
    "VenomRAT",
    "WarzoneRAT",
    "XWorm",
    "XenoRAT",
    "Zloader",
]


@pytest.mark.parametrize(
    "repository_url, extractors, python_minor",
    [
        (
            "https://github.com/jeFF0Falltrades/rat_king_parser",
            ["RKPMACO"],
            10,
        ),
        (
            "https://github.com/apophis133/apophis-YARA-Rules",
            [
                "Pikabot",
                "TrueBot",
                "MetaStealer",
            ],
            8,
        ),
        (
            "https://github.com/kevoreilly/CAPEv2",
            CAPE_EXTRACTORS,
            10,
        ),
    ],
    ids=("jeFF0Falltrades/rat_king_parser", "apophis133/apophis-YARA-Rules", "kevoreilly/CAPEv2"),
)
def test_public_projects(repository_url: str, extractors: list, python_minor: int):
    # Ensure that any changes we make doesn't break usage of public projects
    # which can affect downstream systems using like library (ie. Assemblyline)
    import sys

    from git import Repo
    from tempfile import TemporaryDirectory

    if sys.version_info >= (3, python_minor):
        with TemporaryDirectory() as working_dir:
            project_name = repository_url.rsplit("/", 1)[1]
            extractor_dir = os.path.join(working_dir, project_name)
            Repo.clone_from(repository_url, extractor_dir, depth=1)

            collector = Collector(extractor_dir, create_venv=True)
            assert set(extractors) == set(collector.extractors.keys())

            # Cleanup cached modules to not interfere with later tests
            for module in list(sys.modules.keys()):
                if module not in INIT_MODULES:
                    del sys.modules[module]
    else:
        pytest.skip("Unsupported Python version")


def test_module_confusion():
    # Directories that have the same name as the Python module, shouldn't cause confusion on loading the right module
    collector = Collector(os.path.join(__file__, "../extractors/bob"))
    assert collector.extractors["Bob"]

    collector = Collector(os.path.join(__file__, "../extractors"))
    assert collector.extractors["Bob"]



# Common utilities shared between the MACO collector and configextractor-py
import importlib
import inspect
import json
import os
import pkgutil
import shutil
import subprocess
import sys
import tempfile

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from base64 import b64decode
from glob import glob
from logging import Logger
from sys import executable as python_exe
from typing import Any, Callable, Tuple, Dict
from types import ModuleType

from maco.extractor import Extractor

VENV_DIRECTORY_NAME = ".venv"


class Base64Decoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "__class__" not in obj:
            return obj
        type = obj["__class__"]
        if type == "bytes":
            return b64decode(obj["data"])
        return obj


VENV_SCRIPT = """
import importlib
import json
import os
import sys
import yara

from base64 import b64encode
parent_package_path = "{parent_package_path}"
sys.path.insert(1, parent_package_path)
mod = importlib.import_module("{module_name}")

class Base64Encoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, bytes):
            return dict(__class__="bytes", data=b64encode(o).decode())
        return json.JSONEncoder.default(self, o)
matches = []
if mod.{module_class}.yara_rule:
    matches = yara.compile(source=mod.{module_class}.yara_rule).match("{sample_path}")
result = mod.{module_class}().run(open("{sample_path}", 'rb'), matches=matches)

with open("{output_path}", 'w') as fp:
    if not result:
        json.dump(dict(), fp)
    else:
        try:
            json.dump(result.model_dump(exclude_defaults=True, exclude_none=True), fp, cls=Base64Encoder)
        except AttributeError:
            # venv likely has an older version of Pydantic < 2 installed
            json.dump(result.dict(exclude_defaults=True, exclude_none=True), fp, cls=Base64Encoder)
"""


def maco_extractor_validation(module: ModuleType) -> bool:
    if inspect.isclass(module):
        # 'author' has to be implemented otherwise will raise an exception according to MACO
        return bool(issubclass(module, Extractor) and module.author)
    return False


def maco_extract_rules(module: Extractor) -> bool:
    return module.yara_rule


def create_venv(root_directory: str, logger: Logger, recurse: bool = True):
    # Recursively look for "requirements.txt" or "pyproject.toml" files and create a virtual environment
    for root, _, files in os.walk(root_directory):
        req_files = list({"requirements.txt", "pyproject.toml", "poetry.lock"}.intersection(set(files)))
        if req_files:
            install_command = [
                f"{VENV_DIRECTORY_NAME}/bin/pip",
                "install",
                "-U",
            ]
            venv_path = os.path.join(root, VENV_DIRECTORY_NAME)
            # Create a venv environment if it doesn't already exist
            if not os.path.exists(venv_path):
                logger.info(f"Creating venv at: {venv_path}")
                subprocess.run([python_exe, "-m", "venv", venv_path], capture_output=True)
                # Update pip
                subprocess.run(
                    [f"{VENV_DIRECTORY_NAME}/bin/pip", "install", "--upgrade", "pip"],
                    capture_output=True,
                    cwd=root,
                )
            else:
                logger.info(f"Updating venv at: {venv_path}")

            if "poetry.lock" in req_files:
                # Install dependencies using Poetry
                try:
                    subprocess.run(["poetry", "--version"], capture_output=True, cwd=root)
                    install_command = ["poetry", "install", "--no-root"]
                except FileNotFoundError:
                    logger.warning("Unable to install dependencies using Poetry because it is not installed.")
                    continue

            # Update the pip install command depending on where the dependencies are coming from
            elif "requirements.txt" in req_files:
                # Perform a pip install using the requirements flag
                install_command.extend(["-r", "requirements.txt"])
            elif "pyproject.toml" in req_files:
                # Assume we're dealing with a project directory
                pyproject_command = ["-e", "."]

                # Check to see if there are optional dependencies required
                with open(os.path.join(root, "pyproject.toml"), "rb") as f:
                    parsed_toml_project = tomllib.load(f).get("project", {})
                    for dep_name, dependencies in parsed_toml_project.get("optional-dependencies", {}).items():
                        # Look for the dependency that hints at use of MACO for the extractors
                        if "maco" in " ".join(dependencies):
                            pyproject_command = ["-e", f".[{dep_name}]"]
                            break

                install_command.extend(pyproject_command)

            # Install/Update packages within the venv relative the dependencies extracted
            logger.debug(f"Install command: {' '.join(install_command)}")
            p = subprocess.run(
                install_command,
                cwd=root,
                capture_output=True,
            )
            if p.returncode != 0:
                if b"is being installed using the legacy" in p.stderr:
                    # Ignore these types of errors
                    continue
                logger.error(f"Error installing into venv:\n{p.stderr.decode()}")
                continue
            else:
                logger.debug(f"Installed dependencies into venv:\n{p.stdout}")
                break

        if root == root_directory and not recurse:
            # Limit venv creation to the root directory
            break


def find_extractors(
    parsers_dir: str,
    logger: Logger,
    extractor_module_callback: Callable[[ModuleType, ModuleType, str], bool],
):
    parsers_dir = os.path.abspath(parsers_dir)
    logger.debug("Adding directories within parser directory in case of local dependencies")
    logger.debug(f"Adding {os.path.join(parsers_dir, os.pardir)} to PATH")

    # Specific feature for Assemblyline or environments wanting to run parsers from different sources
    # The goal is to try and introduce package isolation/specification similar to a virtual environment when running parsers
    root_venv = None
    if VENV_DIRECTORY_NAME in os.listdir(parsers_dir):
        root_venv = os.path.join(parsers_dir, VENV_DIRECTORY_NAME)

    # Find extractors (taken from MaCo's Collector class)
    path_parent, foldername = os.path.split(parsers_dir)
    original_dir = parsers_dir
    sys.path.insert(1, path_parent)

    # To avoid module confusion, don't add the directory that has the same name as a Python module within to the PATH
    if f"{foldername}.py" not in os.listdir(parsers_dir):
        sys.path.insert(1, parsers_dir)

    if "src" in os.listdir(parsers_dir):
        # The actual module might be located in the src subdirectory
        # Ref: https://packaging.python.org/en/latest/discussions/src-layout-vs-flat-layout/
        src_path = os.path.join(parsers_dir, "src")
        sys.path.insert(1, src_path)

        # The module to be loaded should be the directory within src
        foldername = [
            d for d in os.listdir(src_path) if os.path.isdir(os.path.join(src_path, d)) and not d.endswith(".egg-info")
        ][0]

    if root_venv:
        # Insert the venv's site-packages into the PATH temporarily to load the module
        for dir in glob(os.path.join(root_venv, "lib/python*/site-packages")):
            sys.path.insert(2, dir)
            break
        mod = importlib.import_module(foldername)
        sys.path.pop(2)
    else:
        mod = importlib.import_module(foldername)

    if mod.__file__ and not mod.__file__.startswith(parsers_dir):
        # Library confused folder name with installed package
        sys.path.remove(path_parent)
        sys.path.remove(parsers_dir)
        parsers_dir = tempfile.TemporaryDirectory().name
        shutil.copytree(original_dir, parsers_dir, dirs_exist_ok=True)

        path_parent, foldername = os.path.split(parsers_dir)
        sys.path.insert(1, path_parent)
        sys.path.insert(1, parsers_dir)
        mod = importlib.import_module(foldername)

    # walk packages in the extractors directory to find all extactors
    for module_path, module_name, ispkg in pkgutil.walk_packages(mod.__path__, mod.__name__ + "."):

        def find_venv(path: str) -> str:
            parent_dir = os.path.dirname(path)
            if VENV_DIRECTORY_NAME in os.listdir(path):
                # venv is in the same directory as the parser
                return os.path.join(path, VENV_DIRECTORY_NAME)
            elif parent_dir == parsers_dir or path == parsers_dir:
                # We made it all the way back to the parser directory
                # Use root venv, if any
                return root_venv
            elif VENV_DIRECTORY_NAME in os.listdir(parent_dir):
                # We found a venv before going back to the root of the parser directory
                # Assume that because it's the closest, it's the most relevant
                return os.path.join(parent_dir, VENV_DIRECTORY_NAME)
            else:
                # Keep searching in the parent directory for a venv
                return find_venv(parent_dir)

        if ispkg:
            # skip __init__.py
            continue

        if module_name.endswith(".setup"):
            # skip setup.py
            continue

        logger.debug(f"Inspecting '{module_name}' for extractors")

        # Local site packages, if any, need to be loaded before attempting to import the module
        parser_venv = find_venv(module_path.path)
        parser_site_packages = None
        if parser_venv:
            for dir in glob(os.path.join(parser_venv, "lib/python*/site-packages")):
                if dir not in sys.path:
                    sys.path.insert(1, dir)
                    break

        try:
            module = importlib.import_module(module_name)
        except BaseException as e:
            # Log if there was an error importing module
            logger.error(f"{module_name}: {e}")
            continue
        finally:
            if parser_site_packages in sys.path:
                sys.path.remove(parser_site_packages)
        # Determine if module contains parsers of a supported framework
        candidates = [module] + [member for _, member in inspect.getmembers(module) if inspect.isclass(member)]
        for member in candidates:
            try:
                if "test" in member.__name__.lower():
                    continue
                if extractor_module_callback(member, module, parser_venv):
                    # If the callback returns with a positive response, we can move onto the next module
                    break

            except TypeError:
                pass
            except Exception as e:
                logger.error(f"{member}: {e}")

    if original_dir != parsers_dir:
        # Correct the paths to the parsers to match metadata changes
        sys.path.remove(path_parent)
        sys.path.remove(parsers_dir)
        path_parent, _ = os.path.split(original_dir)
        sys.path.insert(1, path_parent)
        sys.path.insert(1, original_dir)
        shutil.rmtree(parsers_dir)


def run_in_venv(
    sample_path,
    module,
    module_path,
    venv,
    venv_script=VENV_SCRIPT,
    json_decoder=Base64Decoder,
) -> Dict[str, dict]:
    # Write temporary script in the same directory as extractor to resolve relative imports
    python_exe = os.path.join(venv, "bin", "python")
    dirname = os.path.dirname(module_path)
    with tempfile.NamedTemporaryFile("w", dir=dirname, suffix=".py") as script:
        with tempfile.NamedTemporaryFile() as output:
            module_name = module.__module__
            module_class = module.__name__
            parent_package_path = dirname.rsplit(module_name.split(".", 1)[0], 1)[0]
            root_directory = module_path[:-3].rsplit(module_name.split(".", 1)[1].replace(".", "/"))[0]

            script.write(
                venv_script.format(
                    parent_package_path=parent_package_path,
                    module_name=module_name,
                    module_class=module_class,
                    sample_path=sample_path,
                    output_path=output.name,
                )
            )
            script.flush()
            cwd = root_directory
            custom_module = script.name[:-3].replace(root_directory, "").replace("/", ".")

            if custom_module.startswith("src."):
                # src layout found, which means the actual module content is within 'src' directory
                custom_module = custom_module[4:]
                cwd = os.path.join(cwd, "src")

            proc = subprocess.run(
                [python_exe, "-m", custom_module],
                cwd=cwd,
                capture_output=True,
            )
            try:
                # Load results and return them
                output.seek(0)
                return json.load(output, cls=json_decoder)
            except Exception:
                # If there was an error raised during runtime, then propagate
                delim = f'File "{module_path}"'
                exception = proc.stderr.decode()
                if delim in exception:
                    exception = f"{delim}{exception.split(delim, 1)[1]}"
                raise Exception(exception)



import io
import os

from maco import base_test


class TestBasicLonger(base_test.BaseTest):
    """Test that an extractor containing the name of another extractor works properly."""

    name = "BasicLonger"
    path = os.path.join(__file__, "..")

    def test_run(self):
        ret = self.extract(io.BytesIO(b"BasicLonger"))
        self.assertEqual(ret["family"], "basic_longer")


class TestBasic(base_test.BaseTest):
    """Test that an extractor containing the name of another extractor works properly."""

    name = "Basic"
    path = os.path.join(__file__, "..")

    def test_run(self):
        ret = self.extract(io.BytesIO(b"Basic"))
        self.assertEqual(ret["family"], "basic")



#!/usr/bin/env python3
"""Setup script."""

from setuptools import find_packages, setup

setup(
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_packages(".", exclude=["test", "tests", "extractors"]),
    include_package_data=True,
    install_requires=[
        r.strip() for r in open("requirements.txt", "r") if not r.startswith("#")
    ],
    name="maco",
    description="",
    author="",
    author_email="",
    classifiers=[],
    entry_points={
        "console_scripts": [
            "maco = maco.cli:main",
        ],
    },
)
